
# This directory provides several test cases. Eventually there should be one
# for each module in the kernel. Each one provides a function main(verbose=False)
# which returns True if and only if the test succeeds.

from . import algebraicapproximation
from . import encoding
from . import equippedtriangulation
from . import interval
from . import lamination
from . import lll
from . import matrix
from . import numberfield
from . import permutation
from . import polynomial
from . import triangulation
from . import triangulation3

